import pandas as pd
from pathlib import Path

from gptables import GPWorkbook, GPTable


def produce_workbook(
        filename,
        sheets,
        theme = None,
        cover = None,
        auto_width = False,
        disable_footer_parentheses = False,
        ):
    """
    Produces a GPWorkbook, ready to be written to the specified `.xlsx` file
    using the ``.close()`` method.

    Parameters
    ----------
    filename : str
        path to write final workbook to (an `.xlsx` file)
    sheets : dict
        mapping worksheet labels to gptables.GPTable objects
    theme : gptables.Theme, optional)
        formatting to be applied tot GPTable elements. gptheme is used by
        default
    cover : gptables.Cover, optional
        cover page text. Including this argument will generat a cover page
    auto_width : bool, optional
        indicate if column widths should be automatically determined. False
        by default.
    disable_footer_parentheses : bool, optional
        indicate if addition of parentheses to footer elements should be
        disabled. Note that disabling this decreases machine-readability.
        
    Returns
    -------
    workbook : gptables.GPWorkbook
    """
    if isinstance(filename, Path):
        filename = filename.as_posix()

    wb = GPWorkbook(filename)

    if theme is not None:
        wb.set_theme(theme)

    if cover is not None:
        ws = wb.add_worksheet(cover.cover_label)
        ws.write_cover(cover, sheets, auto_width)
    
    for sheet, gptable in sheets.items():
        ws = wb.add_worksheet(sheet)
        ws.write_gptable(gptable, auto_width, disable_footer_parentheses)
    
    return wb


def write_workbook(
        filename,
        sheets,
        theme = None,
        cover = None,
        auto_width = False,
        disable_footer_parentheses = False,
        ):

    """
    Writes a GPWorkbook to the specified `.xlsx` file.

    This is an alternative main function that will take in data and theme
    information. It calls upon the package to write a formatted `.xlsx`
    file to the specified path.

    Parameters
    ----------
    filename : str
        Path to write final workbook to (an `.xlsx` file)
    sheets : dict
        mapping worksheet labels to ``GPTable`` objects
    theme : gptables.Theme, optional
        formatting to be applied tot GPTable elements. ``gptheme`` is used by
        default
    cover : gptables.Cover, optional
        cover page text. Including this argument will generat a cover page
    auto_width : bool, optional
        indicate if column widths should be automatically determined. False by default.
    disable_footer_parentheses : bool, optional
        indicate if addition of parentheses to footer elements should be
        disabled. Note that disabling this decreases machine-readability.

    Returns
    -------
    None
    """
    wb = produce_workbook(
        filename,
        sheets,
        theme,
        cover,
        auto_width,
        disable_footer_parentheses
        )
    wb.close()


def quick_and_dirty_workbook(
    filename,
    tables,
    theme = None,
    auto_width = True
    ):
    """
    Writes a list of tables to the specified `.xlsx` file, with no metadata.

    This function may be useful for creating simple outputs,
    with no ``title``, ``notes`` or other associated metadata.
    Tables must be ``pandas.DataFrame`` objects with the row indexes
    in the first 1, 2 or 3 columns.

    Parameters
    ----------
    filename : str
        path to write final workbook to (an `.xlsx` file)
    tables : list[pandas.DataFrame]
        ordered tables to be written to file
    theme : gptables.Theme, optional
        formatting to be applied tot GPTable elements. ``gptheme`` is used by
        default
    auto_width : bool, optional
        indicate if column widths should be automatically determined. True by default.

    Returns
    -------
    None
    """
    sheets = dict()
    for table_n in range(len(tables)):
        current_table = tables[table_n]
        if not isinstance(current_table, pd.DataFrame):
            raise TypeError("All tables must be pandas.DataFrame objects")
        
        index_columns = dict()
        index_column_map = {
            0: {2: 0},
            1: {1: 0, 2: 1},
            2: {1: 0, 2: 1, 3: 2}
        }
        cols = current_table.shape[1]
        for n in range(min(cols, 3)):
            if any(current_table.iloc[:, n].apply(lambda x: isinstance(x, str))):
                index_columns.update(index_column_map[n])
            
        sheets["Table " + str(table_n + 1)] = GPTable(
            current_table,
            None,
            None,
            None,
            None,
            index_columns = index_columns
            )

    write_workbook(
        filename = filename,
        sheets = sheets,
        theme = theme,
        auto_width = auto_width,
        disable_footer_parentheses = True
        )



import pytest
import os
from contextlib import redirect_stdout
from pkg_resources import resource_filename
from itertools import chain, combinations

from gptables import Theme
from gptables import gptheme



valid_footer_elements = ["legend", "notes", "annotations", "source"]


def powerset(iterable):
    s = list(iterable)
    return chain.from_iterable(combinations(s, r) for r in range(len(s)+1))

@pytest.fixture()
def empty_theme():
    yield Theme()


class TestCleanInitTheme:
    """
    Test initialisation of the Theme class without config.
    """ 
    @pytest.mark.parametrize("attr", Theme()._format_attributes)
    def test_default_format_attrs(self, attr, empty_theme):  
        """Test Theme attribute default types"""
        exp = {}
        got = getattr(empty_theme, attr)
        assert exp == got
    

    def test_default_other_attrs(self, empty_theme):
        assert empty_theme.footer_order == []
        assert empty_theme.missing_value == None


    def test_print_attributes(self, empty_theme):
        """Test Theme print_attributes()"""
        from io import StringIO

        file_handler = StringIO()
        with redirect_stdout(file_handler):
            empty_theme.print_attributes()
        
        got = file_handler.getvalue()

        exp = (
"""cover_title_format : {}
cover_subtitle_format : {}
cover_text_format : {}
title_format : {}
subtitle_format : {}
scope_format : {}
units_format : {}
column_heading_format : {}
index_1_format : {}
index_2_format : {}
index_3_format : {}
data_format : {}
source_format : {}
legend_format : {}
annotations_format : {}
notes_format : {}
footer_order : []
missing_value : None
"""
                )

        assert got == exp

class TestConfigInitTheme:
    """
    Test initialisation of the Theme class using a config dictionary.
    """
    def test_dict_init(self):
        config = {
            "global":
                {
                 "font_size": 9,
                "font_name": "Arial"
                },

            "cover_title":
                {
                    'font_size': 12,
                    'bold': True
                },
            
            "cover_subtitle":
                {
                    'font_size': 10,
                    'bold': True
                },

            "cover_text": None,

            "title":
                {
                 "bold": True,
                "font_size": 11
                },
     
            "subtitle":
                {"font_size": 10},
            
            "scope": None,
            
            "units":
                {
                "align": "right",
                "italic": True
                },
            
            "column_heading":
                {
                "bold": True,
                "bottom": 1
                },
            
            "index_1":
                {"bold": True},
            
            "index_2": None,
            
            "index_3":
                {"italic": True},
            
            "data": None,
            
            "source":
                {"font_size": 7},
            
            "legend":
                {"font_size": 7},
            
            "annotations":
                {"font_size": 7},
            
            "notes":
                {"font_size": 7},
            
            "footer_order":
                [
                "source",
                "legend",
                "annotations",
                "notes",
                ],
            
            "missing_value": ':'
                }
        got = Theme(config)
        
        exp = gptheme
            
        assert exp == got
        
    def test_file_init(self):
        """
        Test initialisation of Theme using default theme yaml config file.
        """
        config_file = resource_filename(
            "gptables",
            "themes/gptheme.yaml"
            )
        got = Theme(config_file)
        
        exp = gptheme
                
        assert exp == got


class TestFormatValidationTheme:
    """
    Test validation of format dictionaries.
    """
    def test_invalid_attribute_config(self, empty_theme):
        """
        Test that invalid attribute names in config raises a ValueError.
        """
        config = {"potato": {"font_size": 9}}
        with pytest.raises(ValueError):
            empty_theme.apply_config(config)


    def test_invalid_format_label_config(self, empty_theme):
        """
        Test that invalid format labels in config raises a ValueError.
        """
        config = {"notes": {"not_a_format": 5}}
        with pytest.raises(ValueError):
            empty_theme.apply_config(config)
        

    @pytest.mark.parametrize("attr", Theme()._format_attributes)
    def test_invalid_format_label_single_attr(self, attr, empty_theme):
        """
        Test that `validate_single_format` decorator catches invalid format
        labels in individual format dictionaries.
        """
        format_dict = {"not_a_format": 5}
        with pytest.raises(ValueError):
            getattr(empty_theme, "update_" + attr)(format_dict)
    

    @pytest.mark.parametrize("attr", Theme()._format_attributes)
    def test_valid_format_label(self, attr, empty_theme):
        """
        Test that Theme attribute is correctly updated when valid format label
        is used.
        """
        getattr(empty_theme, "update_" + attr)({"font_size": 9})
        
        exp = {"font_size": 9}
        got = getattr(empty_theme, attr)
        
        assert exp == got

    
    def test_valid_format_label_config(self, empty_theme):
        """
        Test that valid format labels in config changes specified format attr
        but not an unrelated attr. Previous bug updated all formats whenever
        one was updated.
        """
        config = {"notes": {"font_size": 5}}
        empty_theme.apply_config(config)
        
        exp = {"font_size": 5}
        got = empty_theme.notes_format
        
        assert exp == got
        
        got2 = empty_theme.title_format
        
        assert {} == got2


class TestOtherValidationTheme:
    """
    Test validation of non-format Theme attributes.
    """
    @pytest.mark.parametrize(
        "format_order",[
            "notes",
            {"annotations": 2},
            1,
            3.14,
            False,
            None
            ]
        )
    def test_invalid_footer_order_type(self, format_order, empty_theme):
        """
        Test that non-list footer_order entries raise a TypeError.
        """
        with pytest.raises(TypeError):
            empty_theme.update_footer_order(format_order)


    @pytest.mark.parametrize(
        "format_order",[
            ["potato"],
            [1],
            [3.14],
            [dict()],
            [[]]
            ]
        )             
    def test_invalid_footer_order_values(self, format_order, empty_theme):
        """
        Test that list footer_order entries containing invalid elements raises
        a ValueError.
        """
        with pytest.raises(ValueError):
            empty_theme.update_footer_order(format_order)
    

    @pytest.mark.parametrize("footer_order", powerset(valid_footer_elements))
    def test_valid_footer_order_values(self, footer_order, empty_theme):
        """
        Test that valid list footer_order entries are used to set
        the corresponding attribute.
        """
        empty_theme.update_footer_order(list(footer_order))
        assert getattr(empty_theme, "footer_order") == list(footer_order)



"""
Iris - Additional Formatting Example
------------------------------------

This example demonstrates additional formatting that is not supported in
the ``gptable.Theme``.

Specific columns, rows and cells of the table elements (indexes, column headings and data)
can be formatted using the ``gptable.GPTable(..., additional_formatting = ...)`` parameter.
This parameter takes a list of dictionaries, allowing you to select as many rows, columns
or cells as you like.

As with all formatting, supported arguments are desribed in the
`XlsxWriter documentation <https://xlsxwriter.readthedocs.io/format.html#format-methods-and-format-properties>`_.

Any formatting not possibly through this means can be achieved using
``XlsxWriter`` `Workbook <https://xlsxwriter.readthedocs.io/workbook.html>`_
and `Worksheet <https://xlsxwriter.readthedocs.io/worksheet.html>`_ functionality.
A ``gptable.GPWorkbook`` object is returned when using the
``gptables.produce_workbook`` API function.
The ``GPWorkbook.worksheets()`` function returns a list of ``GPWorksheet`` objects,
which can also be modified.
"""
import gptables as gpt
import pandas as pd
import numpy as np
import os
from pathlib import Path

## Read data and arrange
parent_dir = Path(__file__).parent

iris_data = pd.read_csv(parent_dir / "iris.csv")

iris_data.rename(
    columns={
        "class": "class",
        "sepal_length": "Sepal Length",
        "petal_length": "Petal Length",
        "petal_width": "Petal Width",
        "sepal_width": "Sepal Width",
    },
    inplace=True,
    )

iris_data["class"] = iris_data.apply(lambda row: row["class"][5:].capitalize(), axis=1)

# Calculate summaries
subtables = []
funcs = [np.mean, np.median]
for func in funcs:
    subtables.append(iris_data.groupby("class").agg(func))
    subtables.append(pd.DataFrame(iris_data.agg(func).rename("All")).T)

iris_summary = pd.concat(subtables)
iris_summary["func"] = ["Mean"] * 4 + ["Median"] * 4

# Reshape
iris_summary = iris_summary.reset_index()
iris_summary = iris_summary.melt(["index", "func"])
iris_summary = iris_summary.pivot_table(
    index=["variable", "func"], columns="index", values="value"
    ).reset_index()

## Define table elements
title = "Iris flower dimensions"
subtitles = [
    "1936 Fisher, R.A; The use of multiple measurements in taxonomic problems",
    "Just another subtitile",
    ]
units = "cm"
scope = "Iris"
index = {1: 0, 2: 1}

## Define additional formatting
# Columns can be references by name or number
# Rows may only be referenced by number
# Column and row numbers refer to the table elements, including indexes and column headings
additional_formatting = [
    {
        "column": {
            "columns": ["Setosa", "Versicolor"],  # str, int or list of either
            "format": {"align": "center"},
            "include_names": False,  # Whether to include column headings (optional)
        }
    },
    {"column": {"columns": [3], "format": {"left": 1}, "include_names": True}},
    {
        "row": {
            "rows": -1,  # Numbers only, but can refer to last row using -1
            "format": {"bottom": 1},  # Underline row
            "include_names": True,  # Whether to include row indexes
        }
    },
    {
        "cell": {
            "cells": (3, 3),  # tuple or list of tuples
            "format": {"font_color": "red"},
        }
    },
    ]

# or just use kwargs
kwargs = {
    "title": title,
    "subtitles": subtitles,
    "units": units,
    "scope": scope,
    "source": None,
    "index_columns": index,
    "additional_formatting": additional_formatting,
    }

## Define our GPTable
iris_table = gpt.GPTable(table=iris_summary, **kwargs)

## Use produce workbook to return GPWorkbook
if __name__ == "__main__":
    output_path = parent_dir / "python_iris_additional_formatting_gptable.xlsx"
    wb = gpt.produce_workbook(
        filename=output_path, sheets={"Iris Flower Dimensions": iris_table}
        )

    # Carry out additional modifications on the GPWorkbook or GPWorksheets
    # This supports all `XlsxWriter` package functionality
    ws = wb.worksheets()[0]
    ws.set_row(0, 30)  # Set the height of the first row

    # Finally use the close method to save the output
    wb.close()
    print("Output written at: ", output_path)




import unittest
from io import StringIO
from collections import namedtuple
import pandas as pd
from pandas.testing import assert_frame_equal, assert_series_equal

import xlsxwriter

import gptables
from gptables.core.wrappers import GPWorkbook
from gptables.core.wrappers import GPWorksheet
from gptables import Theme
from gptables import gptheme
import pytest

Tb = namedtuple("Testbook", "wb ws")

valid_text_elements = [  # Not None
    "This is a string",
    ["More than ", {"italic": True}, "just ", "a string"]
]

test_text_list = [
    "This has a $$reference$$",
    "This one doesn't",
    "Here's another $$one$$"
    ]

exp_text_list = [
    "This has a (1)",
    "This one doesn't",
    "Here's another (2)"
    ]


invalid_text_elements = [
    dict(),
    set(),
    42,
    3.14,
    True
]

@pytest.fixture()
def testbook():
    # See https://github.com/jmcnamara/XlsxWriter/issues/746#issuecomment-685869888
    wb = GPWorkbook(options={'in_memory': True})
    ws = wb.add_worksheet()
    yield Tb(wb, ws)
    wb.fileclosed = 1


class TestGPWorksheetInit:
    """
    Test that default attributes are set when GPWorksheets are created.
    """
    
    def test_subclass(self):
        """
        Test that the GPWorksheet class is a subclass of the XlsxWriter
        Worksheet class.
        """
        assert issubclass(
                        GPWorksheet,
                        xlsxwriter.worksheet.Worksheet
                        )
    

    def test_default_theme_set(self, testbook):
        """
        Test that the default theme (gptheme) is used when no theme is set.
        """
        assert testbook.wb.theme == gptheme
    

    def test_default_gridlines(self, testbook):
        """
        Test that print and screen gridlines are hidden by default.
        """
        assert testbook.ws.print_gridlines == 0
        assert testbook.ws.screen_gridlines == 0
        

    def test_wb_reference(self, testbook):
        """
        Test that GPWorksheets reference their parent GPWorkbook.
        """
        assert testbook.ws._workbook == testbook.wb
    

    @pytest.mark.parametrize("not_a_gptable", [
        dict(),
        set(),
        [],
        1,
        3.14,
        "test_string",
        pd.DataFrame()
    ])
    def test_invalid_write_gptable(self, not_a_gptable, testbook):
        """
        Test that write_gptable() raises a TypeError when argument is not a
        gptables.GPTable object.
        """
        with pytest.raises(TypeError):
            testbook.ws.write_gptable(not_a_gptable)


class TestGPWorksheetWriting:
    """
    Test that additional writing methods correctly write to GPWorksheet object.
    """

    def test__smart_write_str(self, testbook):
        """
        Test that strings are stored in the GPWorksheet as expected.
        """
        testbook.ws._smart_write(0, 0, valid_text_elements[0], {})
        # Strings are stored in a lookup table for efficiency
        got_string = testbook.ws.str_table.string_table
        exp_string = {valid_text_elements[0]: 0}
        assert got_string == exp_string
        
        # String is referenced using a named tuple (string, Format)
        # Here we get first element, which references string lookup location
        got_lookup = testbook.ws.table[0][0][0]
        exp_lookup = 0
        assert got_lookup == exp_lookup


    def test__smart_write_formatted_str(self, testbook):
        testbook.ws._smart_write(1, 2, valid_text_elements[0], {"bold": True})
        # Strings are stored in a lookup table for efficiency
        got_string = testbook.ws.str_table.string_table
        exp_string = {valid_text_elements[0]: 0}
        assert got_string == exp_string
        
        # String is referenced using a named tuple (string, Format)
        # Here we get first element, which references string lookup location
        cell = testbook.ws.table[1][2] 
        got_lookup = cell[0]
        exp_lookup = 0
        assert got_lookup == exp_lookup
        
        format_obj = cell[1]
        assert format_obj.bold
        

    def test__smart_write_rich_text(self, testbook):
        testbook.wb.set_theme(Theme({}))
        
        testbook.ws._smart_write(0, 0, valid_text_elements[1], {})
        # Strings are stored in a lookup table for efficiency
        got_string = testbook.ws.str_table.string_table
        exp_string = {'<r><t xml:space="preserve">More than </t></r><r><rPr><i'
                      '/><sz val="11"/><color theme="1"/><rFont val="Calibri"/'
                      '><family val="2"/><scheme val="minor"/></rPr><t xml:spa'
                      'ce="preserve">just </t></r><r><rPr><sz val="11"/><color'
                      ' theme="1"/><rFont val="Calibri"/><family val="2"/><sch'
                      'eme val="minor"/></rPr><t>a string</t></r>': 0}
        assert got_string == exp_string
        
        # String is referenced using a named tuple (string, Format)
        # Here we get first element, which references string lookup location
        got_lookup = testbook.ws.table[0][0][0]
        exp_lookup = 0
        assert got_lookup == exp_lookup
    


    def test__smart_write_formatted_rich_text(self, testbook):
        testbook.wb.set_theme(Theme({}))
        
        testbook.ws._smart_write(1, 2, valid_text_elements[1], {"bold": True})
        # Strings are stored in a lookup table for efficiency
        got_string = testbook.ws.str_table.string_table
        exp_string = {'<r><t xml:space="preserve">More than </t></r><r><rPr><b'
                      '/><i/><sz val="11"/><color theme="1"/><rFont val="Calib'
                      'ri"/><family val="2"/><scheme val="minor"/></rPr><t xml'
                      ':space="preserve">just </t></r><r><rPr><b/><sz val="11"'
                      '/><color theme="1"/><rFont val="Calibri"/><family val="'
                      '2"/><scheme val="minor"/></rPr><t>a string</t></r>': 0}

        assert got_string == exp_string
        
        # String is referenced using a named tuple (string, Format)
        # Here we get first element, which references string lookup location
        cell = testbook.ws.table[1][2] 
        got_lookup = cell[0]
        exp_lookup = 0
        assert got_lookup == exp_lookup
        
        format_obj = cell[1]
        assert format_obj.bold



class TestGPWorksheetFooterText:
    """
    Test that GPTable footer elements are modified correctly by GPWorksheet
    during write_gptable().
    """


    @pytest.mark.parametrize("text,enclosed", [
        ("", "()"),
        ("This has a $$reference$$", "(This has a $$reference$$)"),
        ("(Already have some)", "((Already have some))")
        ]
    )
    def test__enclose_text_string(self, text, enclosed, testbook):
        """
        Test that strings are correctly flanked with parentheses.
        """
        test_text = testbook.ws._enclose_text(text)
        assert test_text == enclosed


    def test__enclose_text_list(self, testbook):
        """
        Test that rich text is flanked with parentheses.
        """
        test_list = ["An", {"bold": True}, "example", "note"]
        got = testbook.ws._enclose_text(test_list)
        exp = ["(", "An", {"bold": True}, "example", "note", ")"]
        assert got == exp

    @pytest.mark.parametrize("text", test_text_list)
    def test__replace_reference(self, text, testbook):
        """
        Test that references ($$ref$$ style) in strings are replaced with
        numbers, in order of appearance. Also tests replacement in lists.
        """
        got_output = []
        ordered_refs = []

        got_output = [testbook.ws._replace_reference(text, ordered_refs) for text in test_text_list]
    
        exp_refs = ["reference", "one"]
        assert ordered_refs == exp_refs
        assert got_output == exp_text_list


    @pytest.mark.parametrize("text,refs,output",
        zip(test_text_list,
        [["reference"], [], ["one"]],
        ["This has a (1)", "This one doesn't", "Here's another (1)"]
        ))
    def test__replace_reference_in_attr_str(self, text, refs, output, testbook):
        """
        Test that references are replaced in a single string.
        """
        ordered_refs = []
        got_text = testbook.ws._replace_reference_in_attr(
                text,
                ordered_refs
                )

        assert ordered_refs == refs
        assert got_text == output


    def test__replace_reference_in_attr_dict(self, testbook):
        """
        Test that references are replaced in dictionary values, but not keys.
        """
        ordered_refs = []
        test_text_dict = {
                "$$key$$": "This is a value with a $$reference$$",
                "another_key": "Another value",
                "third_key": "$$one$$ more reference"
                }
        got_text = testbook.ws._replace_reference_in_attr(
                test_text_dict,
                ordered_refs
                )
        
        assert ordered_refs == ["reference", "one"]
        
        exp_text_dict = {
                "$$key$$": "This is a value with a (1)",
                "another_key": "Another value",
                "third_key": "(2) more reference"
                }
        
        assert got_text == exp_text_dict



class TestGPWorksheetFormatUpdate:
    """
    Test that GPWorksheet format updating methods work as expected.
    """
    def test__apply_format_dict(self, testbook):
        test = dict()
        format_dict = {"bold": True}
        testbook.ws._apply_format(test, format_dict)
        exp = {"bold": True}
        assert test == exp


    def test__apply_format_series(self, testbook):
        test = pd.Series([{} for n in range(3)])
        format_dict = {"bold": True}
        testbook.ws._apply_format(test, format_dict)
        exp = pd.Series([{"bold": True} for n in range(3)])
        assert_series_equal(test, exp)


    def test__apply_format_dataframe(self, testbook):
        test = pd.DataFrame(columns=[0, 1, 2], index = [0, 1])
        test.iloc[0] = [{} for n in range(3)]
        test.iloc[1] = [{} for n in range(3)]
        
        format_dict = {"bold": True}
        testbook.ws._apply_format(test, format_dict)
        exp = pd.DataFrame(columns=[0, 1, 2], index = [0, 1])
        exp.iloc[0] = [{"bold": True} for n in range(3)]
        exp.iloc[1] = [{"bold": True} for n in range(3)]
        assert_frame_equal(test, exp)


class TestGPWorkbookStatic:

    @pytest.mark.parametrize("input, expected", [
        ("no references", "no references"),
        ("ref at end$$1$$", "ref at end"),
        ("$$1$$ref at start", "ref at start"),
        ("two$$1$$ refs$$2$$", "two refs"),
        ("three$$1$$ refs$$2$$, wow$$3$$", "three refs, wow")
    ])
    def test__strip_annotation_references(self, input, expected):
        assert GPWorksheet._strip_annotation_references(input) == expected
        
        
class TestGPWorkbook:
    """
    Test that GPWorkbook initialisation and methods work as expected.
    """ 
    def test_subclass(self):
        """
        Test that the GPWorkbook class is a subclass of the XlsxWriter
        Workbook class.
        """
        assert issubclass(
                        GPWorkbook,
                        xlsxwriter.Workbook
                        )
    

    def test_default_theme_set(self, testbook):
        """
        Test that the workbook theme is set to gptheme by default.
        """        
        assert testbook.wb.theme == gptheme
    

    def test_valid_set_theme(self, testbook):
        """
        Test that setting a new theme with a Theme object works as expected.
        Essentially, make sure that gptheme is not used.
        """
        theme_config = {"title": {"bold": True}}
        theme = gptables.Theme(theme_config)
        testbook.wb.set_theme(theme)
        
        assert testbook.wb.theme == gptables.Theme(theme_config)
    
    @pytest.mark.parametrize("not_a_theme", [
        dict(),
        set(),
        [],
        1,
        3.14,
        "test_string",
        pd.DataFrame()
    ])
    def test_invalid_set_theme(self, not_a_theme, testbook):
        """
        Test that setting theme with an object that is not a gptables.Theme
        raises a TypeError.
        """
        with pytest.raises(TypeError):
            testbook.wb.set_theme(not_a_theme)



"""
Iris - Quick and Dirty Example
------------------------------

This example demonstrates use of the ``gptables.quick_and_dirty_workbook`` function.
This function is intended for use when only data, and no metadata, is required.
It should **not** be used for production of reference statistical tables for publication.

This function takes a list of ``pandas.DataFrame`` objects and writes each to a
separate sheet of the specified output `.xlsx` file. Formatting is taken from
the default ``gptables.Theme`` - gptheme - unless an alternative theme is provided
using the ``theme`` parameter. Column widths are automatically adjusted,
unless the ``auto_width`` parameter is set to False.

Unlike the other API functions, ``quick_and_dirty_workbook`` does not require
the user to specify table metadata or indicate which columns are row indexes.
Up to 3 row index levels are automatically detected. However, as with the other
API functions, row indices must be columns in the table instead of set as a ``pandas.Index``.
"""
import gptables as gpt
import pandas as pd
import numpy as np
import os
from pathlib import Path

## Read data and arrange
parent_dir = Path(__file__).parent

iris_data = pd.read_csv(parent_dir / "iris.csv")
iris_data = iris_data.loc[:, ["class", "sepal_length", "sepal_width"]]

iris_summary = iris_data.groupby("class").agg(np.mean)
iris_summary.index = [_[5:].capitalize() for _ in iris_summary.index]
iris_summary.rename(
    columns={
        "class":"class",
        "sepal_length":"Mean Sepal Length",
        "sepal_width":"Mean Sepal Width"
    }, 
    inplace=True
    )

# Drop index into table
iris_summary.reset_index(inplace=True)

# Make tables with varying numbers of index levels - up to 3 supported
iris_summary_0 = iris_summary.copy().iloc[:, 1:]

iris_summary_2 = iris_summary.copy()
iris_summary_2.insert(0, "top_index", ["A", "B", "C"])

iris_summary_3 = iris_summary_2.copy()
iris_summary_3.insert(2, "bottom_index", ["x", "y", "z"])


## Write using quick and dirty method
if __name__ == "__main__":
    output_path = parent_dir / "python_iris_quick_and_dirty.xlsx"
    gpt.quick_and_dirty_workbook(
        filename=output_path,
        tables=[iris_summary_0, iris_summary, iris_summary_2, iris_summary_3]
    )
    print("Output written at: ", output_path)




from gptables.core.theme import Theme
from gptables.core.cover import Cover
from gptables.core.gptable import GPTable
from gptables.core.wrappers import GPWorkbook

from gptables.utils.unpickle_themes import gptheme


from gptables.core.api import (
        # API functions
        produce_workbook,
	    write_workbook,
        quick_and_dirty_workbook
        )

__doc__ = """
*******************************
Good Practice Tables (gptables)
*******************************

.. image:: https://travis-ci.org/best-practice-and-impact/gptables.svg?branch=master
    :target: https://travis-ci.org/best-practice-and-impact/gptables
    :alt: Travis build status

.. image:: https://badge.fury.io/py/gptables.svg
    :target: https://badge.fury.io/py/gptables
    :alt: PyPI release


``gptables`` is an opinionated python package for spreadsheet production.
It produces ``.xlsx`` files from your ``pandas`` dataframes or using
``reticulate`` in R. You define the mapping from your data to elements of the
table. It does the rest.

``gptables`` uses the official `guidance on good practice spreadsheets`_
It advocates a strong adherence to the guidance by restricting the range of
operations possible. The default theme ``gptheme`` should accommodate most use
cases. However, the ``Theme`` object allows development of custom themes, where
other formatting is required.

``gptables`` is developed and maintained by the `Best Practice and Impact`_
division of the Office for National Statistics, UK.

.. _`guidance on good practice spreadsheets`: https://gss.civilservice.gov.uk/policy-store/releasing-statistics-in-spreadsheets/

.. _`Best Practice and Impact`: https://gss.civilservice.gov.uk/about-us/support-for-the-gss/



4 Simple Steps
--------------

1. You map your data to the elements of a ``GPTable``.

2. You can define the format of each element with a custom ``Theme``, or simply use the default - gptheme.

3. Optionally design a ``Cover`` page to provide information that relates to all of the tables in your Workbook.

4. You ``write_workbook`` to win.
"""



"""
Cost of Reoffending - Multiple Sheets Example
---------------------------------------------

This example demonstrates how multiple sheets can be defined and written
to a single .xlsx file using ``gptables``.

The elements dictionary below is used to define the metadata for each table,
with the sheet name as it's key. This metadata is used to generate the sheets
dictionary. ``gptables.write_workbook`` uses this dictionary to write each table
to the corresponding sheet.

``gptables.GPTable`` objects can be constructed one by one, but this demonstrates
one way to make this definition concise.
"""
import gptables as gpt
import pandas as pd
import os
from pathlib import Path

## Read data and arrange
parent_dir = Path(__file__).parent
core_data = pd.read_csv(parent_dir / "coreB.csv")

# 3 tables: summary, children and young people, and adults
summ = core_data.loc[:, ["age", "total_cost"]]
summ = summ.groupby("age").sum().reset_index()

# This returns adults first as per the data
ages = core_data.age.unique()

dct = {ages[0]: pd.DataFrame(), ages[1]: pd.DataFrame()}
for key in dct:
    frame = core_data.loc[core_data.age == key, ["reoffence_group", "total_cost"]]
    frame = frame.groupby("reoffence_group").sum().reset_index()
    dct[key] = frame

dct["summary"] = summ


## Define table elements for each table

example = {
    "title": "Cost of Reoffending",
    "subtitles": ["12-month follow-up period for the 2016 offender cohort"],
    "units": "£",
    "scope": "England and Wales",
    "source": "Office for National Statistics",
    }

elements = {"summary": example, "Adults": example, "Children and young people": example}

## Generate a dictionary of sheet names to GPTable objects
## using the elements defined above
sheets = {name: gpt.GPTable(dct[name], **elements[name]) for name in dct}

## Use write_workbook to win!
if __name__ == "__main__":
    output_path = parent_dir / "python_cor_multiple_gptables.xlsx"
    gpt.write_workbook(filename=output_path, sheets=sheets)
    print("Output written at: ", output_path)



from xlsxwriter.format import Format
from gptables.core.gptable import GPTable
import yaml
from functools import wraps

def validate_single_format(f):
    @wraps(f)
    def wrapper(cls, format_dict):
        """
        Decorator to validate that input is a dictionary dictionary.
        """
        if not isinstance(format_dict, dict):
            actual_type = type(format_dict)
            msg = ("Formats must be supplied as a dictionary, not"
                   f"{actual_type}")
            raise ValueError(msg)
            
        for fmt in format_dict.keys():
            cls._validate_format_label(fmt)
        return f(cls, format_dict)
    return wrapper

class Theme:
    """
    A class that defines a set of format attributes for use in xlsxwriter.

    This class associates a dict of format attributes with table elements.

    See XlsxWriter
    `format properties <https://xlsxwriter.readthedocs.io/format.html#format-methods-and-format-properties>`_
    for valid options.

    Attributes
    ----------
    cover_title_format : dict

    cover_subtitle_format : dict

    cover_text_format : dict
    
    title_format : dict

    subtitle_format : dict
    
    scope_format : dict
    
    units_format : dict

    column_heading_format : dict
    
    index_1_format : dict
    
    index_2_format : dict
    
    index_3_format : dict
    
    data_format : dict

    source_format : dict
    
    legend_format : dict
    
    annotations_format : dict
    
    notes_format : dict
    
    footer_order : list
    
    missing_value : None or str
    """

    def __init__(
            self,
            config=None,
            ):
        """
        Initialise theme object.

        Parameters
        ----------
        config : dict or .yaml/.yml file
          theme specification
        """
        ## Formats
        self._format_attributes = [
            "cover_title_format",
            "cover_subtitle_format",
            "cover_text_format",
            "title_format",
            "subtitle_format",
            "scope_format",
            "units_format",
            "column_heading_format",
            "index_1_format",
            "index_2_format",
            "index_3_format",
            "data_format",
            "source_format",
            "legend_format",
            "annotations_format",
            "notes_format"
            ]
        
        for attr in self._format_attributes:
            setattr(self, attr, {})
        
        ## Other attributes
        self.footer_order = []
        self.missing_value = None
        
        # Valid Them format attributes
        self._valid_attrs = [
            x.replace("_format", "")
            for x in self._format_attributes
            ] + ["global"]

        # Valid XlsxWriter Format attributes
        self._valid_format_labels = [
                attr.replace("set_", "")
                for attr in Format().__dir__() 
                if attr.startswith('set_')
                and callable(getattr(Format(), attr))
                ]
            
        if config:
            self.apply_config(config)
    

    @staticmethod
    def _parse_config(config):
        """
        Parse yaml configuration to dictionary.
        """
        if isinstance(config, str):
            if not config.endswith((".yml", ".yaml")):
                raise ValueError("Theme configuration files must be YAML")
            with open(config, "r") as file:
                cfg = yaml.safe_load(file)
                
        elif isinstance(config, dict):
            cfg = config
            
        else:
            raise ValueError("Theme configuration must be a dict or YAML file")
            
        return cfg


    def _validate_config(self, config):
        """
        Assert that format dictionary lower level keys are valid XlsxWriter
        Format attributes.
        """
        for attr in config.keys():
            if attr in self._valid_attrs:
                attr_config = config[attr] or {}
                for fmt in attr_config.keys():
                    self._validate_format_label(fmt)
     

    def _validate_format_label(self, format_name):
        """
        Assert that format is a valid XlsxWriter Format attribute.
        """
        if format_name not in self._valid_format_labels:
            raise ValueError(f"`{format_name}` is not a valid format label")
    

    def apply_config(self, config):
        """
        Update multiple Theme attributes using a YAML or dictionary config.
        This enables extension of build in Themes.
        """
        cfg = self._parse_config(config)
        self._validate_config(cfg)
        
        # Update all when global used
        if "global" in cfg.keys():
            default_format = cfg.pop("global")
            self._update_all_formats(default_format)
        
        # Update with individual methods
        for key, value in cfg.items():
            if key in ["footer_order", "missing_value"]:
                getattr(self, "update_" + key)(value)
            elif key in self._valid_attrs:
                if value is not None:
                    getattr(self, "update_" + key + "_format")(value)
            else:
                raise ValueError(f"`{key}` is not a valid Theme attribute")
    

    def _update_all_formats(self, global_dict):
        """
        Updates all theme attributes with a global format dictionary.
        """
        for attr in self._format_attributes:
            if attr.endswith("_format"):
                getattr(self, "update_" + attr)(global_dict)


    @validate_single_format
    def update_column_heading_format(self, format_dict):
        """
        Update the `column_heading_format` attribute. Where keys already exist,
        existing items are replaced.
        """
        self.column_heading_format.update(format_dict)
    

    @validate_single_format
    def update_index_1_format(self, format_dict):
        """
        Update the `index_1_format` attribute. Where keys already exist, existing
        items are replaced.
        """
        self.index_1_format.update(format_dict)


    @validate_single_format    
    def update_index_2_format(self, format_dict):
        """
        Update the `index_2_format` attribute. Where keys already exist, existing
        items are replaced.
        """
        self.index_2_format.update(format_dict)


    @validate_single_format
    def update_index_3_format(self, format_dict):
        """
        Update the `index_3_format` attribute. Where keys already exist, existing
        items are replaced.
        """
        self.index_3_format.update(format_dict)


    @validate_single_format
    def update_data_format(self, format_dict):
        """
        Update the `data_format` attribute. Where keys already exist,
        existing items are replaced.
        """
        self.data_format.update(format_dict)


    @validate_single_format
    def update_cover_title_format(self, format_dict):
        """
        Update the `cover_title_format` attribute. Where keys already exist, existing
        items are replaced.
        """
        self.cover_title_format.update(format_dict)
    
    
    @validate_single_format
    def update_cover_subtitle_format(self, format_dict):
        """
        Update the `cover_subtitle_format` attribute. Where keys already exist, existing
        items are replaced.
        """
        self.cover_subtitle_format.update(format_dict)


    @validate_single_format
    def update_cover_text_format(self, format_dict):
        """
        Update the `cover_text_format` attribute. Where keys already exist, existing
        items are replaced.
        """
        self.cover_text_format.update(format_dict)

    @validate_single_format
    def update_title_format(self, format_dict):
        """
        Update the `title_format` attribute. Where keys already exist, existing
        items are replaced.
        """
        self.title_format.update(format_dict)


    @validate_single_format
    def update_subtitle_format(self, format_dict):
        """
        Update the `subtitle_format` attribute. Where keys already exist,
        existing items are replaced.
        """
        self.subtitle_format.update(format_dict)


    @validate_single_format    
    def update_scope_format(self, format_dict):
        """
        Update the `scope_format` attribute. Where keys already exist, existing
        items are replaced.
        """
        self.scope_format.update(format_dict)


    @validate_single_format
    def update_location_format(self, format_dict):
        """
        Update the `location_format` attribute. Where keys already exist,
        existing items are replaced.
        """
        self.location_format.update(format_dict)


    @validate_single_format
    def update_units_format(self, format_dict):
        """
        Update the `units_format` attribute. Where keys already exist, existing
        items are replaced.
        """
        self.units_format.update(format_dict)


    @validate_single_format
    def update_source_format(self, format_dict):
        """
        Update the `source_format` attribute. Where keys already exist,
        existing items are replaced.
        """
        self.source_format.update(format_dict)


    @validate_single_format
    def update_legend_format(self, format_dict):
        """
        Update the `legend_format` attribute. Where keys already exist,
        existing items are replaced.
        """
        self.legend_format.update(format_dict)


    @validate_single_format
    def update_annotations_format(self, format_dict):
        """
        Update the `annotations_format` attribute. Where keys already exist,
        existing items are replaced.
        """
        self.annotations_format.update(format_dict)


    @validate_single_format
    def update_notes_format(self, format_dict):
        """
        Update the `notes_format` attribute. Where keys already exist, existing
        items are replaced.
        """
        self.notes_format.update(format_dict)


    def update_footer_order(self, order_list):
        """
        Update the `footer_order` attribute. Overrides existing order.
        """
        if not isinstance(order_list, list):
            msg = ("`footer_order` must be a list of footer element names")
            raise TypeError(msg)

        valid_elements = ["source", "legend", "notes", "annotations"]
        if not all(element in valid_elements for element in order_list):
            msg = (f"`footer_order` elements must be in {valid_elements}")
            raise ValueError(msg)
        self.footer_order = order_list


    def update_missing_value(self, missing_val_text):
        """
        Update the `missing_value` attribute. Overrides existing string.
        """
        GPTable._validate_text(missing_val_text, "missing_value")

        self.missing_value = missing_val_text


    def print_attributes(self):
        """
        Print all current format attributes and values to the console.
        """
        obj_attr = [
                attr for attr in self.__dir__()
                if not attr.startswith('_')
                and not callable(getattr(self, attr))
                ]
        for attr in obj_attr:
            print(attr, ":", getattr(self, attr))
            

    def __eq__(self, other):
        """
        Comparison operator, for testing.
        """
        # don't attempt to compare against unrelated types
        if not isinstance(other, Theme):
            return False
        
        obj_attr = [
                attr for attr in self.__dir__()
                if not attr.startswith('_')
                and not callable(getattr(self, attr))
                ]
        return all([
                getattr(self, attr) == getattr(other, attr)
                for attr in obj_attr
                ])
        


import pandas as pd
import pytest
from pandas.testing import assert_frame_equal
from contextlib import contextmanager


from gptables import GPTable


# TODO: These should be stored in GPTable
gptable_text_attrs = ["title", "scope", "units", "source"]

gptable_list_text_attrs = ["subtitles", "legend", "notes"]


valid_index_columns = [
    {},
    {1: "one"},
    {1: "one", 2: "two"},
    {1: "one", 2: "two", 3: "three"}
    ]

valid_text_elements = [
    "This is a string",
    ["This is ", {"bold": True}, "rich", "text"],
    None
]

invalid_text_elements = [
    dict(),
    set(),
    42,
    3.14,
    True
]


@contextmanager
def does_not_raise():
    yield


@pytest.fixture(scope="function")
def create_gptable_with_kwargs():

    def generate_gptable(format_dict=None):
        base_gptable = {
            "table": pd.DataFrame(),
            "title": "",
            "scope": "",
            "units": "",
            "source": "",
            "index_columns": {}  # Override default, as no columns in table
            }
        if format_dict is not None:
            base_gptable.update(format_dict)
        return GPTable(**base_gptable)

    return generate_gptable


def test_init_defaults(create_gptable_with_kwargs):
    """
    Test that given a minimal input, default attributes are correct types.
    """
    empty_gptable = create_gptable_with_kwargs()

    # Required args
    assert empty_gptable.title == ""
    assert empty_gptable.scope == ""
    assert empty_gptable.units == ""
    assert empty_gptable.source == ""
    assert_frame_equal(
            empty_gptable.table, pd.DataFrame().reset_index(drop=True)
            )
    
    # Optional args
    assert empty_gptable.index_columns == {}
    assert empty_gptable.subtitles == []
    assert empty_gptable.legend == []
    assert empty_gptable.annotations == {}
    assert empty_gptable.notes == []
    assert empty_gptable.additional_formatting == []
    
    # Other
    assert empty_gptable.index_levels == 0
    assert empty_gptable._column_headings == set()
    assert empty_gptable._VALID_INDEX_LEVELS == [1, 2, 3]
    

class TestAttrValidationGPTable:
    @pytest.mark.parametrize("level", [4, 0, -1])
    def test_invalid_index_level(self, level, create_gptable_with_kwargs):
        """
        Test that GPTable index_columns raises error when an index level is
        invalid.
        """
        with pytest.raises(ValueError):
            create_gptable_with_kwargs({
                "table": pd.DataFrame(columns=["one", "two", "three"]),
                # Valid column index, but invalid level
                "index_columns": {level: 0}
                })
                

    @pytest.mark.parametrize("idx,expectation", [
        (0, does_not_raise()),
        (1, does_not_raise()),
        (3, pytest.raises(ValueError)),
        (-1, pytest.raises(ValueError))
    ])
    def test_invalid_column_index(self, idx, expectation, create_gptable_with_kwargs):
        """
        Test that GPTable index_columns raises error when a column index number
        is invalid.
        """
        with expectation:
            create_gptable_with_kwargs({
                "table": pd.DataFrame(columns=["one", "two", "three"]),  
                # Valid index level, but non existent column
                "index_columns": {1: idx}
                })
            

    @pytest.mark.parametrize("index_cols,col_headings", zip(
        valid_index_columns, [
            {0, 1, 2, 3},
            {1, 2, 3},
            {2, 3},
            {3}
            ]
        ))
    def test_set_column_index(self, index_cols, col_headings, create_gptable_with_kwargs):
        """
        Test that setting GPTable index_columns with valid column index works
        as expected.
        """
        gptable = create_gptable_with_kwargs({
            "table": pd.DataFrame(columns=["one", "two", "three", "four"]),
            "index_columns": index_cols
            })
        assert gptable.index_columns == index_cols
        assert gptable.index_levels == len(index_cols)
        assert gptable._column_headings == col_headings
        

    @pytest.mark.parametrize("not_a_table", [1, "?", [1,3,5,6]])
    def test_invalid_table_type(self, not_a_table, create_gptable_with_kwargs):
        """
        Test that setting GPTable table to object that is not a
        pandas.DataFrame raises an error.
        """
        with pytest.raises(TypeError):
            create_gptable_with_kwargs({"table": not_a_table})
    

    @pytest.mark.parametrize("attr", gptable_text_attrs)
    @pytest.mark.parametrize("not_text", invalid_text_elements)
    def test_invalid_text_in_str_attrs(self, attr, not_text, create_gptable_with_kwargs):
        """
        Test that setting an invalid GPTable text types raises a TypeError for
        each attribute that holds a string.
        """
        with pytest.raises(TypeError):
            create_gptable_with_kwargs({attr: not_text})
    

    @pytest.mark.parametrize("attr", gptable_text_attrs)
    @pytest.mark.parametrize("text", valid_text_elements)
    def test_valid_text_in_str_attrs(self, attr, text, create_gptable_with_kwargs):
        """
        Test that setting valid GPTable text elements works as expected. Test
        strings and list containing strings and format dicts (rich text).
        Also test that None is allowed.
        """
        gptable = create_gptable_with_kwargs({attr: text})
        assert getattr(gptable, attr) == text
    

    @pytest.mark.parametrize("attr", gptable_list_text_attrs)
    @pytest.mark.parametrize("text", invalid_text_elements)
    def test_invalid_text_in_list_attrs(self, attr, text, create_gptable_with_kwargs):
        """
        Test that setting list of invalid text elements to GPTable list
        parameters raises a TypeError.
        """
        text = [text, text]
        with pytest.raises(TypeError):
            create_gptable_with_kwargs({attr: text})


    @pytest.mark.parametrize("attr", gptable_list_text_attrs)
    @pytest.mark.parametrize("text", valid_text_elements)
    def test_valid_text_in_list_attrs(self, attr, text, create_gptable_with_kwargs):
        """
        Test that setting list of valid text elements to GPTable list
        parameters works as expected.
        """
        if text is not None:
            text = [text, text]
        gptable = create_gptable_with_kwargs({attr: text})
        if text is not None:
            assert getattr(gptable, attr) == text
        else:
            assert getattr(gptable, attr) == []


    @pytest.mark.parametrize("reference", [42, (0, 0), None])
    def test_invalid_annotations_keys(self, reference, create_gptable_with_kwargs):
        """
        Test that setting annotations keys that are not strings raise a
        TypeError.
        """    
        with pytest.raises(TypeError):
            create_gptable_with_kwargs({
                "annotations": {reference: "valid value"}
                })
    

    @pytest.mark.parametrize("reference", ["1", "spam"])
    def test_valid_annotations_keys(self, reference, create_gptable_with_kwargs):
        """
        Test that setting str annotation keys works.
        """
        gptable = create_gptable_with_kwargs({
            "annotations": {reference: "valid value"}
            })
        assert getattr(gptable, "annotations") == {reference: "valid value"}
    

    @pytest.mark.parametrize("text", invalid_text_elements)
    def test_invalid_annotations_values(self, text, create_gptable_with_kwargs):
        """
        Test that setting annotations values that are not valid text elements
        raises a TypeError.
        """
        with pytest.raises(TypeError):
            create_gptable_with_kwargs({
            "annotations": {"valid_key": text}
            })
    

    @pytest.mark.parametrize("text", valid_text_elements)
    def test_valid_annotations_values(self, text, create_gptable_with_kwargs):
        """
        Test that setting annotations values that valid text elements works as
        expected.
        """        
        gptable = create_gptable_with_kwargs({
            "annotations": {"valid_key": text}
            })
        assert getattr(gptable, "annotations") == {"valid_key": text}
    

    @pytest.mark.parametrize("key", invalid_text_elements[2:] + ["invalid_key"])
    def test_invalid_additional_format_keys(self, key, create_gptable_with_kwargs):
        """
        Test that adding additional formatting with an invalid key raises an
        error.
        """
        with pytest.raises(ValueError):
            create_gptable_with_kwargs({
            "additional_formatting": [{key: {"format": {"bold": True}}}]
            })
        

    @pytest.mark.parametrize("key", ["cell", "row", "column"])
    def test_valid_additional_format_keys(self, key, create_gptable_with_kwargs):
        """
        Test that adding additional formatting with a valid key (column, row or
        cell) works as expected.
        """
        gptable = create_gptable_with_kwargs({
            "additional_formatting": [
                    {key: {"format": {"bold": True}}}]
            })
        assert getattr(gptable, "additional_formatting") == [{key: {"format": {"bold": True}}}]
    

    @pytest.mark.parametrize("format_label", invalid_text_elements[2:] + ["not_a_format"])
    def test_invalid_additional_format_labels(self, format_label, create_gptable_with_kwargs):
        """
        Test that adding additional formatting with a format parameter that is
        not supported by XlsxWriter raises an error.
        """
        with pytest.raises(ValueError):
            create_gptable_with_kwargs({
                "additional_formatting": 
                    [
                        {"cell":
                            {"format": {format_label: True},
                                "cells": (0, 0)
                                }
                        }
                    ]
                })
    

    @pytest.mark.parametrize(
        "format_dict",[
            {"bold": True},
            {"font_size": 17},
            {"align": "center"},
            {"font_color": "red"},
            {"bottom": 1}
            ])
    def test_valid_additional_format_labels(self, format_dict, create_gptable_with_kwargs):
        """
        Test that adding additional formatting with a format parameter that is
         supported by XlsxWriter works as expected.
        """
        additional_formatting = [
                    {"cell":
                        {"format": format_dict,
                            "cells": (0, 0)
                            }
                        }
                    ]
        gptable = create_gptable_with_kwargs({
             "additional_formatting": additional_formatting
            })
        assert getattr(gptable, "additional_formatting") == additional_formatting


@pytest.mark.parametrize("index_cols", valid_index_columns)
class TestIndirectAttrs:
    """
    Test that non-formatting create_gptable_with_kwargs attributes are indirectly set correctly.
    """
    def test_index_levels_set(self, index_cols, create_gptable_with_kwargs):
        """
        Test that number of index levels are set, when one, two or three
        indexes are used.
        """
        gptable = create_gptable_with_kwargs({
            "table": pd.DataFrame(
                columns = [
                    "one",
                    "two",
                    "three",
                    "four"
                    ]),
            "index_columns": index_cols
            })
        assert getattr(gptable, "index_levels") == len(index_cols)
        

    def test_column_headings_set(self, index_cols, create_gptable_with_kwargs):
        """
        Test that non-index columns are set as column headings.
        """
        gptable = create_gptable_with_kwargs({
            "table": pd.DataFrame(
                columns = [
                    "one",
                    "two",
                    "three",
                    "four"
                    ]),
            "index_columns": index_cols
            })
        
        # Expect all column numbers that have no index level assigned
        exp = set(range(4)) - set(range(len(index_cols)))
        
        assert getattr(gptable, "_column_headings") == exp



import pandas as pd
from xlsxwriter.format import Format

class GPTable:
    """
    A Good Practice Table. Stores a table and metadata for writing a table
    to excel.
    
    Attributes
    ----------
    table : pandas.DataFrame
        table to be written to an Excel workbook
    title : str
        description of the table
    subtitles : list
        subtitles as a list of strings
    scope : str
        description of scope/basis of data in table
    units : str or dict
        units used in all (str) or each (dict) column of `table`
    legend : list
        descriptions of special notation used in `table`
    annotations : dict
        notes that are referenced in header or table elements (excluding data)
    notes : list
        notes that are not referenced
    index_columns : dict
        mapping an index level to a 0-indexed column as {level: column}.
        Default is a level two index in the first column ({2: 0}).
    additional_formatting : dict
        table-specific formatting for columns, rows or individual cells
    """

    def __init__(self,
                 table,
                 title,
                 scope,
                 units,
                 source,
                 subtitles=[],
                 legend=[],
                 annotations={},
                 notes=[],
                 index_columns={2:0},
                 additional_formatting=[]
                 ):
        
        # Attributes
        self.title = None
        self.subtitles = []
        
        self.scope = None
        self.units = None  # str or {units (str):column index (int)} dict
        
        self._VALID_INDEX_LEVELS = [1, 2, 3]
        self.index_levels = 0
        self.index_columns = {}  # {index level (int): column index (int)}
        self._column_headings = set() # Non-index column headings
        self.table = pd.DataFrame()
        
        self.source = None
        self.legend = []
        self.annotations = {}
        self.notes = []
        
        self.additional_formatting = []
        
        # Valid format labels from XlsxWriter
        self._valid_format_labels = [
                attr.replace("set_", "")
                for attr in Format().__dir__() 
                if attr.startswith('set_')
                and callable(getattr(Format(), attr))
                ]
        
        # Call methods to set attributes        
        self.set_title(title)
        self.set_subtitles(subtitles)
        self.set_scope(scope)
        self.set_table(table, index_columns, units)
        self.set_source(source)
        self.set_legend(legend)
        self.set_annotations(annotations)
        self.set_notes(notes)
        self.set_additional_formatting(additional_formatting)
        

    def set_table(self, new_table, new_index_columns = None, new_units = None):
        """
        Set the `table`, `index_columns` and 'units' attributes. Overwrites
        existing values for these attributes.
        """
        if not isinstance(new_table, pd.DataFrame):
            raise TypeError("`table` must be a pandas DataFrame")
        
        default_index = pd.Index(range(new_table.shape[0]))
        if not all(new_table.index == default_index) and not new_table.empty:
            msg = ("`table` index must not contain index data. It can be reset"
                   " before adding to a GPTable (see DataFrame.reset_index())."
                   " Please ensure that index data is stored in the first 1-3"
                   " columns of `table` and is indicated in `index_columns`.")
            raise ValueError(msg)

        self.table = new_table.reset_index(drop=True)

        if new_index_columns is None:
            new_index_columns = self.index_columns
        self.set_index_columns(new_index_columns)
        if new_units is None:
            new_units = self.units
        self.set_units(new_units)
        

    def set_index_columns(self, new_index_columns):
        """
        Set the `index_columns` attribute. Overwrites any existing values.
        A dict must be supplied. This dict should map index level to a
        single 0-indexed column number. All other columns will be considered
        as data columns.
        """
        if isinstance(new_index_columns, dict):
            # Check if levels and values are valid
            valid_levels = all(level in self._VALID_INDEX_LEVELS for level in new_index_columns.keys())
            if not valid_levels:
                msg = ("`index_columns` dictionary keys must be valid index"
                       f" levels: {self._VALID_INDEX_LEVELS}")
                raise ValueError(msg)
            
            if not all(isinstance(col, int) for col in new_index_columns.values()):
                # Convert col name to numeric index
                for key, value in new_index_columns.items():
                    col_iloc = self.table.columns.get_loc(value)
                    new_index_columns.update({key: col_iloc})
                    
            column_indexes = [col for col in new_index_columns.values()]
                
            valid_columns = all(self._valid_column_index(col) for col in column_indexes)
            if not valid_columns:
                msg = ("Out of range - `index_columns` dictionary values must"
                       "be valid, 0-indexed column numbers")
                raise ValueError(msg)
            
            self.index_levels = len(new_index_columns.keys())
            
            self.index_columns = new_index_columns
            self._set_column_headings()
        else:
            msg = ("`index_columns` must be a dict mapping a valid index level"
                   " to a 0-indexed column number")
            raise ValueError(msg)
    

    def _valid_column_index(self, column_index):
        """
        Check if `column_index` is valid, given the `table` shape.
        """
        return column_index in range(self.table.shape[1])
        

    def _set_column_headings(self):
        """
        Sets the `column_headings` attribute to the set of column indexes that
        are not assigned to `index_columns`.
        """
        index_cols = set(self.index_columns.values())
        self._column_headings = {x for x in range(self.table.shape[1])} - index_cols
    

    def set_title(self, new_title):
        """
        Set the `title` attribute.
        """
        self._validate_text(new_title, "title")
        self.title = new_title
    

    def add_subtitle(self, new_subtitle):
        """
        Add a single subtitle to the existing list of `subtitles`.
        """
        self._validate_text(new_subtitle, "subtitles")
        self.subtitles.append(new_subtitle)
    

    def set_subtitles(self, new_subtitles, overwrite=True):
        """
        Set a list of subtitles to the `subtitles` attribute. Overwrites
        existing ist of subtitles by default. If `overwrite` is False, new list
        is appended to existing list of subtitles.
        """
        if new_subtitles is None:
            new_subtitles = []

        if not isinstance(new_subtitles, (list)):
            msg =("`subtitles` must be provided as a list containing strings"
                  " and/or lists of strings and format dictionaries"
                  " (rich text)")
            raise TypeError(msg)

        for text in new_subtitles:
            self._validate_text(text, "subtitles")
            
        if overwrite:
            self.subtitles = new_subtitles
        else:
            self.subtitles += new_subtitles
            

    def set_scope(self, new_scope):
        """
        Set the `scope` attribute.
        """
        self._validate_text(new_scope, "scope")
        self.scope = new_scope        


    def set_units(self, new_units):
        """
        Set the `units` attribute using the supplied str, list or dictionary.
        Units supplied as a list must match the length of column headings,
        excluding index columns. Units as a dict should be in the format
        {column: units_text}. Column can be column name or 0-indexed column
        number in `table`. Index columns cannot have units.
        """            
        if isinstance(new_units, str) or new_units is None:
            self._validate_text(new_units, "units")
        elif isinstance(new_units, list):
            self._validate_text(new_units, "units")
            rich_text =  any(type(_) == dict for _ in new_units)
            if len(new_units) != len(self._column_headings) and not rich_text:
                msg = ("length of `units` list must match the number of"
                       " non-index columns in the `table`")
                raise ValueError(msg)
        elif isinstance(new_units, dict) and len(new_units) > 0:
            units_list = [None for _ in range(len(self._column_headings))]
            for key, value in new_units.items():
                self._validate_text(value, "units")
                if not isinstance(key, int):
                    iloc = self.table.columns.get_loc(key)
                else:
                    iloc = key
                units_list[iloc - self.index_levels] = value
            new_units = units_list

        else:
            msg = ("`units` attribute must be a string, list or dictionary"
                   " ({column: units_text})")
            raise TypeError(msg)
            
        self.units = new_units


    def set_source(self, new_source):
        """
        Set the source attribute to the specified str.
        """
        self._validate_text(new_source, "source")
            
        self.source = new_source
    

    def add_legend(self, new_legend):
        """
        Add a single legend entry to the existing `legend` list.
        """
        self._validate_text(new_legend, "legend")
        self.subtitles.append(new_legend)
    

    def set_legend(self, new_legend, overwrite=True):
        """
        Set a list of legend entries to the `legend` attribute. Overwrites
        existing legend entries by default. If overwrite is False, new entries 
        are appended to the `legend` list.
        """
        if new_legend is None:
            self.legend = []
            return
        if not isinstance(new_legend, list):
            msg = ("`legend` must be provided as a list of text elements")
            raise TypeError(msg)
        for text in new_legend:
            self._validate_text(text, "legend")
        
        if overwrite:
            self.legend = new_legend
        else:
            self.legend += new_legend
            

    def add_annotation(self, new_annotation):
        """
        Add one or more annotations to the existing `annotations` dictionary.
        """
        if not isinstance(new_annotation, dict):
            raise TypeError("`annotations` entries must be dictionaries")
        for text in new_annotation.values():
            self._validate_text(text, "annotations")
        self.annotations.update(new_annotation)
    

    def set_annotations(self, new_annotations, overwrite=True):
        """
        Set a list of notes to the `annotations` attribute. Overwrites existing
        `annotations` dict by default. If overwrite is False, new entries are
        used to update the `annotations` dict.
        """
        if not isinstance(new_annotations, dict):
            msg = ("annotations must be provided as a dictionary of"
                   " {reference: note}")
            raise TypeError(msg)
        
        if not all(isinstance(key, str) for key in new_annotations.keys()):
            raise TypeError("`annotations` keys must be strings")
        
        for text in new_annotations.values():
            self._validate_text(text, "annotations")
            
        if overwrite:
            self.annotations = new_annotations
        else:
            self.annotations.update(new_annotations)
            

    def add_note(self, new_note):
        """
        Add a single note to the existing `notes` list.
        """
        self._validate_text(new_note, "notes")
        self.notes.append(new_note)
    

    def set_notes(self, new_notes, overwrite=True):
        """
        Set a list of notes to the `notes` attribute. Overwrites existing
        `notes` list by default.If overwrite is False, new entries are
        appended to the `notes` list.
        """
        if new_notes is None:
            self.notes = []
            return
        if not isinstance(new_notes, list):
            msg = ("`notes` must be a list of text elements")
            raise TypeError(msg)

        for text in new_notes:
            self._validate_text(text, "notes")
            
        if overwrite:
            self.notes = new_notes
        else:
            self.notes += new_notes


    def set_additional_formatting(self, new_formatting):
        """
        Set a dictionary of additional formatting to be applied to this table.
        """
        if not isinstance(new_formatting, list):
            msg = ("`additional_formatting` must be a list of dictionaries")
            raise TypeError(msg)
        keys = [key for item in new_formatting for key in item.keys()]
        for key in keys:
            if key not in ["column", "row", "cell"]:
                msg = (f"`{key}` is not a supported format type. Please use"
                       " `column`, `row` or `cell`")
                raise ValueError(msg)
        
        self._validate_format_labels(new_formatting)
            
        self.additional_formatting = new_formatting
    

    def _validate_format_labels(self, format_list):
        """
        Validate that format labels are valid property of XlsxWriter Format.
        """
        labels = [label
                  for item in format_list
                  for key in item.keys()
                  for label in item[key]["format"]
                  ]
        for label in labels:
            if label not in self._valid_format_labels:
                msg = (f"`{label}` is not a valid XlsxWriter Format property")
                raise ValueError(msg)
    
    
    @staticmethod
    def _validate_text(obj, attr):
        """
        Validate that an object contains valid text elements. These are either
        strings or list of strings and dictionaries.
        """
        if isinstance(obj, str) or obj is None:
            return None
        
        msg = (f"{attr} text should be provided as strings or lists of"
                   f" strings and dictionaries (rich-text). {type(obj)} are"
                   " not valid text elements.")
        if isinstance(obj, list):
            for element in obj:
                if not isinstance(element, (str, dict)):
                    msg = (f"{attr} text should be provided as strings or"
                           " lists of strings and dictionaries (rich-text)."
                           f" {type(element)} are not valid rich text"
                           " elements.")
                    raise TypeError(msg)
        else:
            msg = (f"{attr} text should be provided as strings or lists of"
                   f" strings and dictionaries (rich-text). {type(obj)} are"
                   " not valid text elements.")
            raise TypeError(msg)



import pickle
from pkg_resources import resource_filename

class ThemeUnpickler(pickle.Unpickler):
    """
    Points the unpickler to the Theme class. Allows unpickling for package
    init.
    """
    def find_class(self, module, name):
        if name =="Theme":
            from gptables.core.theme import Theme
            return Theme
        return super().find_class(module, name)

gptheme = ThemeUnpickler(
        open(
            resource_filename("gptables", "/theme_pickles/gptheme.pickle"),
            "rb"
            )
        ).load()



import os
import re
import pandas as pd
import numpy as np
from copy import deepcopy

from xlsxwriter.workbook import Workbook
from xlsxwriter.worksheet import Worksheet

from .theme import Theme
from .cover import Cover
from .gptable import GPTable
from gptables.utils.unpickle_themes import gptheme


class GPWorksheet(Worksheet):
    """
    Wrapper for an XlsxWriter Worksheet object. Provides a method for writing
    a good practice table (GPTable) to a Worksheet.
    """
    def write_cover(self, cover, sheets, auto_width):
        """
        Write a cover page to the Worksheet. Uses text from a Cover object and
        details of the Workbook contents.

        Parameters
        ----------
        cover : gptables.Cover
            object containing cover sheet text
        sheets : dict
            mapping worksheet labels to gptables.GPTable objects
        """
        theme = self.theme
        pos = [0, 0]

        pos = self._write_element(pos, cover.title, theme.cover_title_format)
        pos[0] += 1

        if cover.intro is not None:
            pos = self._write_element(pos, "Introductory information", theme.cover_subtitle_format)
            pos = self._write_element_list(pos, cover.intro, theme.cover_text_format)
            pos[0] += 1

        if sheets:
            pos = self._write_element(pos, "Contents", theme.cover_subtitle_format)
            for sheet, gptable in sheets.items():
                pos = self._write_hyperlinked_toc_entry(pos, sheet)
                        
                title = self._strip_annotation_references(gptable.title)
                pos = self._write_element(pos, title, theme.cover_text_format)

                if cover.additional_elements is not None:
                    for element in cover.additional_elements:
                        content = getattr(gptable, element)        
                        if element in ["subtitles", "notes"]:
                            content = [self._strip_annotation_references(element) for element in content]
                            pos = self._write_element_list(pos, content, theme.cover_text_format)
                        else:
                            content = self._strip_annotation_references(content)
                            pos = self._write_element(pos, content, theme.cover_text_format)
                pos[1] = 0
            pos[0] += 1

        if cover.about is not None:
            pos = self._write_element(pos, "About these data", theme.cover_subtitle_format)
            pos = self._write_element_list(pos, cover.about, theme.cover_text_format)
            pos[0] += 1

        if cover.contact is not None:
            pos = self._write_element(pos, "Contact", theme.cover_subtitle_format)
            pos = self._write_element_list(pos, cover.contact, theme.cover_text_format)
            pos[0] += 1


        if sheets and auto_width:
            max_link_len = max([len(key) for key in sheets.keys()])
            first_col_width = self._excel_string_width(
                max_link_len,
                theme.cover_text_format.get("font_size") or 10
                )        
            self._set_column_widths([first_col_width])
        

    def write_gptable(self, gptable, auto_width, disable_footer_parentheses):
        """
        Write data from a GPTable object to the worksheet using the workbook
        Theme object for formatting.
        
        Parameters
        ----------
        gptable : gptables.GPTable
            object containing elements of the gptable to be written to the
            Worksheet
        
        Returns
        -------
        None
        """
        if not isinstance(gptable, GPTable):
            raise TypeError("`gptable` must be a gptables.GPTable object")
        
        gptable = deepcopy(gptable)

        theme = self.theme

        # Write each GPTable element using appropriate Theme attr
        pos = [0, 0]

        self._reference_annotations(gptable)

        pos = self._write_element(
                pos,
                gptable.title,
                theme.title_format
                )

        pos = self._write_element_list(
                pos,
                gptable.subtitles,
                theme.subtitle_format
                )

        pos = self._write_table_elements(
                pos,
                gptable,
                auto_width
                )

        if not disable_footer_parentheses:
            self._enclose_footer_elements(gptable)

        footer = theme.footer_order
        for element in footer:
            pos = getattr(self, "_write_" + element)(
                    pos,
                    getattr(gptable, element),
                    getattr(theme, element + "_format")
                    )

    @staticmethod
    def _strip_annotation_references(text):
        """
        Strip annotation references (as $$ $$) from a str or list text element.
        """
        pattern = r"\$\$.*?\$\$"
        if isinstance(text, str):
            no_annotations = re.sub(pattern, "", text)
        elif isinstance(text, list):
            no_annotations = [
                re.sub(pattern, "", part)
                if isinstance(part, str) else part
                for part in text
                ]
        
        return no_annotations

    def _reference_annotations(self, gptable):
        """
        Replace note references with numbered references. Acts on `title`,
        `subtitles`, `table` and `notes` attributes of a GPTable. References 
        are numbered from top left of spreadsheet, working across each row.
        
        Parameters
        ----------
        gptable : gptables.GPTable
            object containing data with references to notes

        Returns
        -------
        None
        """
        elements = [
                "title",
                "subtitles",
                "scope",
                "units",
                "legend"
                ]
        # Store annotation references in order detected
        ordered_refs = []
        
        # Loop through elements, replacing references in strings
        for attr in elements:
            attr_current = getattr(gptable, attr)
            setattr(
                    gptable,
                    attr,
                    self._replace_reference_in_attr(
                            attr_current,
                            ordered_refs
                            )
                    )
        self._reference_table_annotations(gptable, ordered_refs)

        new_annotations = {}
        # Add to dict in order
        for n in range(len(ordered_refs)):
            try:
                new_annotations.update(
                        {n + 1: gptable.annotations[ordered_refs[n]]}
                        )
            except KeyError:
                msg = (f"`{ordered_refs[n]}` has been referenced, but is not"
                       " defined in GPTable.annotations")
                raise KeyError(msg)
        # Warn if all annotations not referenced
        annotations_diff = len(gptable.annotations) - len(new_annotations)
        if annotations_diff:
            output_file = os.path.basename(self._workbook.filename)
            msg =(f"Warning: {annotations_diff} annotations have not been"
                  f" referenced in {output_file}. These annotations are not"
                  " displayed. Use `notes` for notes without references.") 
            print(msg)
        # Replace old notes refs
        gptable.annotations = new_annotations
        

    def _reference_table_annotations(self, gptable, ordered_refs):
         """
         Reference annotations in the table column headings and index columns.
         """
         table = getattr(gptable, "table")
         
         table.columns = self._replace_reference_in_attr(
                 [x for x in table.columns],
                 ordered_refs
                 )
         
         index_columns = gptable.index_columns.values()
         
         for col in index_columns:
             table.iloc[:, col] = table.iloc[:, col].apply(
                     lambda x: self._replace_reference_in_attr(x, ordered_refs)
                     )

         setattr(gptable, "table", table)


    def _replace_reference_in_attr(self, data, ordered_refs):
        """
        Replaces references in a string or list/dict of strings. Works
        recursively on list elements and dict values. Other types are returned
        without modification. Updates `ordered_refs` with newly detected
        references.
        
        Parameters
        ----------
        data : any type
            object containing strings to replace references in
        ordered_refs : list
            list of references used so far. New references will be added to
            this list in order of detection

        Returns
        -------
        string : str
            input string with references replaced with numerical reference (n),
            where n is the order of appearance in the resulting document
        """
        if isinstance(data, str):
            data = self._replace_reference(data, ordered_refs)
        if isinstance(data, list):
            for n in range(len(data)):
                data[n] = self._replace_reference_in_attr(
                        data[n],
                        ordered_refs
                        )
        if isinstance(data, dict):
            for key in data.keys():
                data[key] = self._replace_reference_in_attr(
                        data[key],
                        ordered_refs
                        )

        return data
    

    @staticmethod
    def _replace_reference(string, ordered_refs):
        """
        Given a single string, record occurrences of new references (denoted by
        flanking dollar signs [$$reference$$]) and replace with number
        reference reflecting order of detection.
        
        Parameters
        ----------
        string : str
            the string to replace references within
        ordered_refs : list
            list of references used so far. New references will be added to
            this list in order of detection

        Returns
        -------
        string : str
            input string with references replaced with numerical reference (n),
            where n is the order of appearence in the resulting document
        """
        text_refs = re.findall(r"[$]{2}.*?[$]{2}", string)
        dict_refs = [w.replace("$", "") for w in text_refs]
        for n in range(len(dict_refs)):
            if dict_refs[n] not in ordered_refs:
                ordered_refs.append(dict_refs[n])
            num_ref = "(" + str(ordered_refs.index(dict_refs[n]) + 1) + ")"
            string = string.replace(text_refs[n], num_ref)

        return string


    def _enclose_footer_elements(self, gptable):
        """
        Flank text footer elements with parentheses.
        """
        gptable.source = self._enclose_text(gptable.source)
        gptable.notes = [self._enclose_text(note) for note in gptable.notes]
        gptable.legend = [
            self._enclose_text(symbol) for symbol in gptable.legend
            ]
        gptable.annotations = dict(
            [("(" + str(k), v + ")") for k, v in gptable.annotations.items()]
            )


    @staticmethod
    def _enclose_text(element):
        """
        Enclose text within parentheses. Handles strings and lists
        (rich strings).
        """
        if isinstance(element, str):
            return  "(" + element + ")"
        elif isinstance(element, list):
            return ["("] + element + [")"]


    def _write_element(self, pos, element, format_dict):
        """
        Write a single text element of a GPTable to the GPWorksheet.
        
        Parameters
        ----------
        element : str or list
            the string or list of rich string elements to be written
        format_dict : dict
            format to be applied to string
        pos : list
            the position of the worksheet cell to write the element to

        Returns
        -------
        pos : list
            new position to write next element from
        """
        if element:
            self._smart_write(*pos, element, format_dict)
            pos[0] += 1
        
        return pos       


    def _write_element_list(self, pos, element_list, format_dict):
        """
        Writes a list of elements row-wise.
        
        Parameters
        ----------
        element_list : list
            list of strings or nested list of rich string elements to write,
            one per row
        format_dict : dict
            format to be applied to string
        pos : list
            the position of the worksheet cell to write the elements to

        Returns
        -------
        pos: list
            new position to write next element from
        """
        if element_list:
            for element in element_list:
                pos = self._write_element(pos, element, format_dict)
        
        return pos


    def _write_hyperlinked_toc_entry(self, pos, sheet_name):
        """
        Write a table of contents entry. Includes a hyperlink to the sheet
        in the first column. Then data for that sheet in the second column.

        Parameters
        ----------
        pos : list
            the position of the worksheet cell to write the elements to
        sheet_name : str
            name of sheet to hyperlink to

        Returns
        -------
        pos: list
            new position to write next element from
        """
        theme = self.theme

        link = f"internal:'{sheet_name}'!A1"
        hyperlink_format = deepcopy(theme.cover_text_format)
        hyperlink_format.update({"underline": True, "font_color": "blue"})
        self._smart_write(
            *pos,
            link,
            hyperlink_format,
            sheet_name
            )        

        return [pos[0] , pos[1] + 1]

    def _write_source(self, pos, element, format_dict):
        """
        Alias for writting footer elements by name.
        """
        return self._write_element(pos, element, format_dict)


    def _write_legend(self, pos, element_list, format_dict):
        """
        Alias for writting footer elements by name.
        """
        return self._write_element_list(pos, element_list, format_dict)


    def _write_notes(self, pos, element_list, format_dict):
        """
        Alias for writting footer elements by name.
        """
        return self._write_element_list(pos, element_list, format_dict)


    def _write_annotations(self, pos, annotations_dict, format_dict):
        """
        Writes a list of ordered annotations row-wise.
        
        Parameters
        ----------
        notes_dict : dict
            note associate with each references, as {reference: note}
        format_dict : dict
            format to be applied to string
        pos : list
            the position of the worksheet cell to write the elements to

        Returns
        -------
        pos: list
            new position to write next element from
        """
        for ref, annotation in annotations_dict.items():
            element = f"{ref}: {annotation}"
            pos = self._write_element(pos, element, format_dict)

        return pos


    def _write_table_elements(self, pos, gptable, auto_width):
        """
        Writes the table, scope and units elements of a GPTable. Uses the
        Workbook Theme, plus any additional formatting associated with the
        GPTable. Also replaces `np.nan` with the missing value marker.
        
        Parameters
        ----------
        gptable : gptables.GPTable
            object containing the table and additional formatting data
        pos : list
            the position of the worksheet cell to write the units to
        auto_width : bool
            select if column widths should be determined automatically using
            length of text in index and columns

        Returns
        -------
        pos : list
            new position to write next element from
        """
        # Get theme
        theme = self.theme
        
        # Write scope
        scope = gptable.scope
        self._smart_write(
                *pos,
                scope,
                theme.scope_format
                )
        
        # Write units above each col heading
        pos[1] += gptable.index_levels
        n_cols = len(gptable._column_headings)
        units = gptable.units
        if isinstance(units, str):
            pos[1] += n_cols - 1
            self._smart_write(
                *pos,
                units,
                theme.units_format
                )
            pos[1] += 1

        elif isinstance(units, list):
            for n in range(n_cols):
                self._smart_write(
                    *pos,
                    units[n],
                    theme.units_format
                    )
                pos[1] += 1
        
        # Reset position to left col on next row
        if (units is not None) and (scope is not None):
            pos[0] += 1
        pos[1] = 0
        
        ## Create data array
        index_levels = gptable.index_levels
        index_columns = [col for col in gptable.index_columns.values()]
        data = pd.DataFrame(gptable.table, copy=True)
            
        # Create row containing column headings
        data.loc[-1] = data.columns
        data.index = data.index + 1
        data.sort_index(inplace=True)
        data.iloc[0, index_columns] = ""  # Delete index col headings
        
        ## Create formats array
        # pandas.DataFrame did NOT want to hold dictionaries, so be wary
        formats = pd.DataFrame().reindex_like(data)
        dict_row = [{} for n in range(formats.shape[1])]
        for row in range(formats.shape[0]):
            dict_row = [{} for n in range(formats.shape[1])]
            formats.iloc[row] = dict_row
        
        
        ## Handle missing values
        missing_marker = theme.missing_value

        if data.isna().values.any():
            if missing_marker is not None:
                # Super inefficient format update loop
                # Only run if align not set for data
                if not "align" in theme.data_format.keys():
                    rows, cols = data.shape
                    for row in range(rows):
                        for col in range(cols):
                            if pd.isna(data.iloc[row, col]):
                                (formats.iloc[row, col]
                                .update({"align":"center"})
                                )

                data.fillna(missing_marker, inplace=True)
                gptable.legend.append(f"{missing_marker} not available")
            else:
                msg = ("`Theme.missing_marker` must be assigned if values are"
                       " missing within GPTable.")
                raise ValueError(msg)
        
        
        ## Add Theme formatting to formats dataframe
        self._apply_format(
                formats.iloc[0, index_levels:],
                theme.column_heading_format
                )
        
        self._apply_format(
                formats.iloc[1:, index_levels:],
                theme.data_format
                )
        
        index_level_formats = [
                theme.index_1_format,
                theme.index_2_format,
                theme.index_3_format
                ]
        for level, col in gptable.index_columns.items():
            self._apply_format(
                formats.iloc[1:, col],
                index_level_formats[level - 1]  # Account for 0-indexing
                )
        
        ## Add additional table-specific formatting from GPTable
        self._apply_additional_formatting(
                formats,
                gptable.additional_formatting,
                gptable.index_levels
                )
        
        ## Write table
        pos = self._write_array(pos, data, formats)

        ## Set columns widths
        if auto_width:
            widths = self._calculate_column_widths(data, formats)
            self._set_column_widths(widths)
        
        return pos

    
    def _apply_additional_formatting(
            self,
            formats_table,
            additional_formatting,
            index_levels
            ):
        """
        Apply row, column and cell formatting to dataframe of formats.
        """
        for item in additional_formatting:
            fmt_type = list(item.keys())[0]
            format_desc = item[fmt_type]
            
            if fmt_type == "cell":
                formatting = format_desc["format"]
                cell_ilocs = format_desc["cells"]
                if isinstance(cell_ilocs, tuple):
                    cell_ilocs = [cell_ilocs]
                for row, col in cell_ilocs:
                    formats_table_slice = formats_table.iloc[row, col]
                    
                    self._apply_format(
                        formats_table_slice,
                        formatting
                        )
                return None
            
            if fmt_type == "column":
                cols_iloc = [
                        formats_table.columns.get_loc(col)
                        if isinstance(col, str)
                        else col
                        for col in format_desc["columns"]
                        ]
                row_start = 0
                if "include_names" in format_desc.keys():
                    row_start = 0 if format_desc["include_names"] else 1
    
                formats_table_slice = formats_table.iloc[row_start:, cols_iloc]
                formatting = format_desc["format"]

            elif fmt_type == "row":
                rows_iloc = format_desc["rows"]
                col_start = 0
                if "include_names" in format_desc.keys():
                    col_start = 0 if format_desc["include_names"] else index_levels
    
                formats_table_slice = formats_table.iloc[rows_iloc, col_start:]
                formatting = format_desc["format"]
            
            self._apply_format(
                formats_table_slice,
                formatting
                )


    def _write_array(self, pos, data, formats):
        """
        Write a two-dimensional array to the current Worksheet, starting from
        the specified position.
        
        Parameters
        ----------
        data : pandas.DataFrame
            array of data to be written to Worksheet
        formats : pandas.DataFrame
            array of dictionaries that specify the formatting to be applied
            to each cell of data
        pos : list
            the position of the top left cell to start writing the array from
            
        Returns
        -------
        pos : list
            new position to write next element from
        """
        if data.shape != formats.shape:
            raise ValueError("data and formats arrays must be of equal shape")
        
        rows, cols = data.shape
        for row in range(rows):
            for col in range(cols):
                cell_data = data.iloc[row, col]
                cell_format_dict = formats.iloc[row, col]
                
                self._smart_write(
                        pos[0] + row,
                        pos[1] + col,
                        cell_data,
                        cell_format_dict
                        )
        
        pos = [pos[0] + rows, 0]
        
        return pos
        

    def _smart_write(self, row, col, data, format_dict, *args):
        """
        Depending on the input data, this function will write rich strings or
        use the standard `write()` method. For rich strings, the base format is
        merged with each rich format supplied within data.
        
        Parameters
        ----------
        row : int
            0-indexed row of cell to write to
        col : int
            0-indexed column of cell to write to
        data : str or list
            Simple string to be written with `format_dict` formatting. Or a
            list of alternating string and dictionary objects. Dictionaries
            specify additional formatting to be applied to the following string
            in the list.
        format_dict : dict
            Dictionary containing base format for the string.
            
        Returns
        -------
        None
        """
        wb = self._workbook  # Reference to Workbook that contains sheet
        if isinstance(data, list):
            # At this point, any list should be a rich-text element
            data_with_formats = []
            for item in data:
                # Convert dicts to Format (with merge onto base format)
                if isinstance(item, dict):
                    rich_format = format_dict.copy()
                    rich_format.update(item) 
                    data_with_formats.append(wb.add_format(rich_format))
                else:
                    data_with_formats.append(item)
            if len(data) > 3:
                data_with_formats.insert(-1, wb.add_format(format_dict))
            
            self.write_rich_string(row,
                                   col,
                                   *data_with_formats,
                                   wb.add_format(format_dict),
                                   *args
                                   )
        else:
            # Write handles all other write types dynamically
            self.write(
                    row,
                    col,
                    data,
                    wb.add_format(format_dict),
                    *args
                    )


    @staticmethod
    def _apply_format(format_table_slice, format_dict):
        """
        Update all cells of a given dataframe slice with the format
        dictionary. Handles dict, series or dataframes.
        """
        if isinstance(format_table_slice, pd.Series):
            (format_table_slice
             .apply(lambda d: d.update(format_dict))
             )
        elif isinstance(format_table_slice, pd.DataFrame):
            # Vectorised for 2D
            (format_table_slice
             .apply(np.vectorize(lambda d: d.update(format_dict)))
             )
        elif isinstance(format_table_slice, dict):
            format_table_slice.update(format_dict)


    def _set_column_widths(self, widths):
        """
        Set the column widths using a list of widths.
        """
        for col_number in range(len(widths)):
            self.set_column(
                col_number,
                col_number,
                widths[col_number]
            )


    def _calculate_column_widths(self, table, formats_table):
        """
        Calculate Excel column widths using maximum length of strings
        and the maximum font size in each column of the data table.

        Parameters
        ----------
        table : pd.DataFrame
            data table to calculate widths from
        formats_table: pd.DataFrame
            formats table to retrieve font size from

        Returns 
        -------
        col_widths : list
            width to apply to Excel columns
        """
        cols = table.shape[1]
        max_lengths = [
            table.iloc[:, col].apply(lambda x: len(x)
            if isinstance(x, str) else 0).max()
            for col in range(cols)
            ]

        max_font_sizes = [
            formats_table.iloc[:, col]
            .apply(lambda x: x.get("font_size") or 10).max()
            for col in range(cols)
            ]

        col_widths = [
            self._excel_string_width(l, f)
            for l, f in zip(max_lengths, max_font_sizes)
            ]
        return col_widths

        
    @staticmethod
    def _excel_string_width(string_len, font_size):
        """
        Calculate the rough length of a string in Excel character units.
        This crude estimate does not account for font name or other font format
        (e.g. wrapping).
        
        Parameters
        ----------
        string_len : int
            length of string to calculate width in Excel for
        font_size : int
            size of font
        
        Returns 
        -------
        excel_width : float
            width of equivalent string in Excel
        """    
        if string_len == 0:
            excel_width = 0
        else:
            excel_width = string_len * ((font_size * 0.12) - 0.09)
        
        return excel_width



class GPWorkbook(Workbook):
    """
    Wrapper for and XlsxWriter Workbook object. The Worksheets class has been
    replaced by an alternative with a method for writting GPTable objects.
    """

    def __init__(self, filename=None, options={}):
        super(GPWorkbook, self).__init__(filename=filename, options=options)
        self.theme = None
        # Set default theme
        self.set_theme(gptheme)
        

    def add_worksheet(self, name=None):
        """
        Overwrite add_worksheet() to create a GPWorksheet object.
        
        Parameters
        ----------
        name : str (optional)
            name of the the worksheet to be created
        
        Returns
        -------
        worksheet : gptables.GPWorksheet
            a worksheet object, which supports writing of GPTable objects
        """
        worksheet = super(GPWorkbook, self).add_worksheet(name, GPWorksheet)
        worksheet.theme = self.theme
        worksheet._workbook = self  # Create reference to wb, for formatting
        worksheet.hide_gridlines(2)
        return worksheet


    def set_theme(self, theme):
        """
        Sets the theme for all GPTable objects written to the Workbook.
        
        Parameters
        ----------
        theme : gptables.Theme
            a Theme object containing the formatting to be applied to GPTable
            objects written to Worksheets within this Workbook
        
        Returns
        -------
        None
        """
        if not isinstance(theme, Theme):
            raise TypeError(f"`theme` must be a gptables.Theme object, not: {type(theme)}")
        self.theme = theme



from dataclasses import dataclass
from typing import List

@dataclass
class Cover():
    """
    dataclass for storing cover sheet text.

    Attributes
    ----------
    title : str
        cover page title
    intro : List[str, dict], optional
        introductory text
    about : List[str, dict], optional
        about/notes text
    contact : List[str, dict], optional
        contact details text
    cover_label : str 
        cover page tab label, defaults to Cover
    additional_elements : List[str], optional
        additional GPTable elements to display in the contents table. Allowed
        elements are "subtitles", "scope", "source" and "notes".
    """
    title: str
    intro: List = None
    about: List = None
    contact: List = None
    cover_label: str = "Cover"
    additional_elements: List = None

    def __post_init__(self):
        if self.additional_elements is not None:
            valid_elements = ["subtitles", "scope", "source", "notes"]
            if not all(element in valid_elements for element in self.additional_elements):
                msg = ("Cover `additional_elements` list can only contain"
                    "'subtitles', 'scope', 'source' and 'notes'")
                raise ValueError(msg)



"""
Iris - Cover Page
-----------------

This example demonstrates use of the ``gptables.Cover`` class to create a cover page.

A gptables cover pages contain a range of custom text elements, along with a hyperlinked table of contents.
Text elements are defined as a ``gptables.Cover`` instance, which is passed to the ``cover`` parameter of ``gptables.write_worbook()`` or ``gptables.produce_worbook()``.
In this example, we have also set ``auto_width`` to ``True``.
This automatically determines the width of the first column on the cover sheet, as well as all columns of the tables of the workbook.
"""
import gptables as gpt
import pandas as pd
import numpy as np
import os
from pathlib import Path
from copy import deepcopy

## Read data and arrange
parent_dir = Path(__file__).parent

iris_data = pd.read_csv(parent_dir / "iris.csv")
iris_data = iris_data.loc[:, ["class", "sepal_length", "sepal_width"]]

iris_summary = iris_data.groupby("class").agg(np.mean)
iris_summary.index = [_[5:].capitalize() for _ in iris_summary.index]
iris_summary.rename(
    columns={
        "class": "class",
        "sepal_length": "Mean Sepal Length",
        "sepal_width": "Mean Sepal Width",
    },
    inplace=True,
)

# Drop index into table
iris_summary.reset_index(inplace=True)

# Insert NA to demonstrate missing value representation
iris_summary.iloc[1, 1] = np.nan

## Define table elements
title = ["Mean", {"italic": True}, " Iris", "$$note2$$ sepal dimensions"]
subtitles = [
    "1936 Fisher, R.A; The use of multiple measurements in taxonomic problems$$note1$$",
    "Just another subtitile",
    ]
units = "cm"
scope = "Iris"
source = "Source: Office for Iris Statistics"
index = {2: 0}  # Column 0 is a level 2 index
annotations = {
    "note1": "I've got 99 problems and taxonomy is one.",
    "note2": "Goo Goo Dolls, 1998.",
    }
notes = ["This note hath no reference."]

# or use kwargs to pass these to the appropriate parameters
kwargs = {
    "title": title,
    "subtitles": subtitles,
    "units": units,
    "scope": scope,
    "source": source,
    "index_columns": index,
    "annotations": annotations,
    "notes": notes,
    }

## Define our GPTable
iris_table = gpt.GPTable(table=iris_summary, **kwargs)

iris_table_copy = deepcopy(iris_table)
iris_table_copy.set_title("A copy of the first sheet$$note2$$")

cover = gpt.Cover(
    cover_label="Notes",
    title="A Worbook containing good practice tables",
    intro=["This is some introductory information", "And some more"],
    about=["Even more info about my data", "And a little more"],
    contact=["John Doe", "Tel: 345345345"],
    additional_elements=["subtitles", "scope", "source", "notes"]
    )

## Use write_workbook to win!
if __name__ == "__main__":
    output_path = parent_dir / "python_iris_cover_gptable.xlsx"
    gpt.write_workbook(
        filename=output_path,
        sheets={
            "Iris Flower Dimensions": iris_table,
            "Copy of Iris Flower Dimensions": iris_table_copy,
        },
        cover=cover,
        auto_width=True,
    )
    print("Output written at: ", output_path)



"""
gptables example using the Titanic dataset.
@author: Ramiz Farishta
"""

import gptables as gpt
import pandas as pd
import numpy as np
import os
from pathlib import Path

parent_dir = Path(__file__).parent
titanic = pd.read_csv(parent_dir / "titanic.csv")

titanic["Child"] = np.where(titanic["Age"] < 18, 1, 0)

sex_survived = titanic.groupby("Sex")["Survived"].sum()
sex_pclass = titanic.groupby("Sex")["Pclass"].apply(lambda x: x.mode().iloc[0])
sex_child = titanic.groupby("Sex")["Child"].sum()
sex_fare = titanic.groupby("Sex")["Fare"].mean()

titanic_analysis = pd.DataFrame(
    {
        "Survived": sex_survived,
        "Pclass": sex_pclass,
        "Child": sex_child,
        "Fare": sex_fare,
    }
).reset_index()

title = "Titanic$$note2$$ analysis by sex"
subtitles = ["Derived from a Kaggle competition dataset$$note1$$"]
units = {"Survived": "sum$$note3$$", "Pclass": "mode", "Child": "sum", "Sex": "mean"}
scope = "Titanic$$note2$$"
source = "Source: Kaggle"
index = {2: 0}
annotations = {
    "note1": "www.kaggle.com/titanic",
    "note2": "Celine Dion.",
    "note3": "Total count.",
    }
notes = ["This note hath no reference."]

kwargs = {
    "title": title,
    "subtitles": subtitles,
    "units": units,
    "scope": scope,
    "source": source,
    "index_columns": index,
    "annotations": annotations,
    "notes": notes,
    }

# define our GPTable
titanic_table = gpt.GPTable(table=titanic_analysis, **kwargs)

if __name__ is "__main__":
    output_path = parent_dir / "python_titanic_gptable.xlsx"
    gpt.write_workbook(
        filename=output_path, sheets={"titanic analysis by sex": titanic_table}
        )
    print("Output written at: ", output_path)



from setuptools import setup, find_packages

def read_requirements(file):
    with open(file) as f:
        return f.read().splitlines()

def read_file(file):
   with open(file) as f:
        return f.read()
    
long_description = read_file("README.rst")
version = read_file("VERSION")
requires = read_requirements("requirements.txt")

docs_requires = [
    "sphinx>=2",
    "sphinx_rtd_theme"
    ] + requires

test_requires = [
    "coverage",
    "pytest>=3.6,<4",
    "pytest-cov"
] + docs_requires

setup(
    name = 'gptables',
    version = version,
    author = 'David Foster',
    author_email = 'david.foster@ons.gov.uk',
    url = 'https://best-practice-and-impact.github.io/gptables/',
    keywords="reproducible tables excel xlsxwriter reproducible-analytical-pipelines",
    description = 'Simplifying good practice in statistical tables.',
    long_description_content_type = "text/x-rst",
    long_description = long_description,
    license = "MIT license",
    packages = find_packages(exclude=["test"]),
    include_package_data = True,
    install_requires = requires,
    extras_require = {
        "docs": docs_requires,
        "testing": test_requires
    },
    classifiers = [
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent"
    ],
    python_requires = '>=3.6',
)



import gptables.core.theme
import pickle
import glob
from os.path import abspath, dirname, join, basename, splitext

def pickle_themes():
    """
    Utility function for updating theme pickles.
    """
    package_dir = dirname(dirname(dirname(abspath(__file__))))
    theme_configs = glob.glob(join(
            package_dir,
            "gptables",
            "themes",
            "*.yaml")
    )

    pickled_output_dir = join(package_dir, "gptables", "theme_pickles")

    for cfg in theme_configs:
        file, ext = splitext(basename(cfg))
        out_file = join(pickled_output_dir, (file + ".pickle"))

        theme = gptables.core.theme.Theme(cfg)

        pickle.dump(
            theme,
            open(out_file, "wb")
            )

if __name__ == "__main__":
    pickle_themes()



"""
Iris - Minimal Example
----------------------

This example demonstrates use of the ``gptables.write_workbook`` function.
This API function is designed for production of consistently structured and formatted tables.

Summary statistics from the classic iris dataset are used to build a ``gptables.GPTable``
object. Elements of metadata are provided to the corresponding parameters of the class.
Where you wish to provide no metadata in required parameters, use ``None``.

Table formatting can be defined as a ``gptable.Theme``, which is passed to the API functions
 using the ``theme`` parameter. Or you can reply on our default - gptheme.
"""
import gptables as gpt
import pandas as pd
import numpy as np
from pathlib import Path

## Read data and arrange
parent_dir = Path(__file__).parent

iris_data = pd.read_csv(parent_dir / "iris.csv")
iris_data = iris_data.loc[:, ["class", "sepal_length", "sepal_width"]]

iris_summary = iris_data.groupby("class").agg(np.mean)
iris_summary.index = [_[5:].capitalize() for _ in iris_summary.index]
iris_summary.rename(
    columns={
        "class": "class",
        "sepal_length": "Mean Sepal Length",
        "sepal_width": "Mean Sepal Width",
    },
    inplace=True,
    )

# Drop index into table
iris_summary.reset_index(inplace=True)

# Insert NA to demonstrate missing value representation
iris_summary.iloc[1, 1] = np.nan

## Define table elements
title = ["Mean", {"italic": True}, " Iris", "$$note2$$ sepal dimensions"]
subtitles = [
    "1936 Fisher, R.A; The use of multiple measurements in taxonomic problems$$note1$$",
    "Just another subtitile",
    ]
units = "cm"
scope = "Iris"
source = "Source: Office for Iris Statistics"
index = {2: 0}  # Column 0 is a level 2 index
annotations = {
    "note1": "I've got 99 problems and taxonomy is one.",
    "note2": "Goo Goo Dolls, 1998.",
    }
notes = ["This note hath no reference."]

# or use kwargs to pass these to the appropriate parameters
kwargs = {
    "title": title,
    "subtitles": subtitles,
    "units": units,
    "scope": scope,
    "source": source,
    "index_columns": index,
    "annotations": annotations,
    "notes": notes,
    }

## Define our GPTable
iris_table = gpt.GPTable(table=iris_summary, **kwargs)

## Use write_workbook to win!
if __name__ == "__main__":
    output_path = parent_dir / "python_iris_gptable.xlsx"
    gpt.write_workbook(
        filename=output_path, sheets={"Iris Flower Dimensions": iris_table}
        )
    print("Output written at: ", output_path)
